
using System.ComponentModel.Design;
using Microsoft.VisualStudio.Modeling.Shell;
using System;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Diagrams;
using System.Collections.Generic;

namespace DslFactory.Utilities.Commands
{
 /// <summary>
 /// Standard DSL Commands.
 /// Due to the architecture of the CommandSet - namely the visilibity of its properties, it is more 
 /// a code snippet source than a real library class.
 /// Indeed, the provided methods for the OnStatusX(), and OnMenuX() should be held by the
 /// CommandSet derived class for you DSL, and you should copy and paste each region (such as Copy, etc ...)
 /// to your commandset partial, and use the AddXCommand() in your GetMenuCommands() override
 /// </summary>
 public abstract class DSLStandardCommands : CommandSet
 {
  /// <summary>
  /// Just for the compilation !
  /// </summary>
  /// <param name="serviceProvider"></param>
  protected DSLStandardCommands(IServiceProvider serviceProvider)
   : base(serviceProvider)
  {
  }

  #region Copy
  /// <summary>
  /// Add the support for the standard Copy Command
  /// </summary>
  /// <param name="commands">CommandList, we want to add the Copy command to</param>
  /// <typeparam name="T">A ModelElement derived type which is used to restrict the kind of model element
  /// that can be copied to the clipboard</typeparam>
  public void AddCopyCommand<T>(List<MenuCommand> commands) where T : ModelElement
  {
   commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusCopy),
                                             new EventHandler(OnMenuCopy<T>),
                                             StandardCommands.Copy));
  }

  /// <summary>
  /// Status for the Copy command
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnStatusCopy(object sender, EventArgs e)
  {
   MenuCommand command = sender as MenuCommand;
   command.Visible = true;
   command.Enabled = this.CurrentSelection.Count > 0;
  }

  /// <summary>
  /// Copy command
  /// </summary>
  /// <typeparam name="T"></typeparam>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnMenuCopy<T>(object sender, EventArgs e) where T : ModelElement
  {
   Diagram diagram = this.CurrentDocView.CurrentDiagram;
   bool foundSome = false;
   ElementGroup elementGroup = new ElementGroup(diagram.Partition);
   foreach (object o in this.CurrentSelection)
   {
    // Pick out shapes representing Component model elements.
    ShapeElement element = o as ShapeElement;
    if (element != null && element.ModelElement != null
                        && element.ModelElement is T)
    {
     // add the element and its embedded children to the group
     elementGroup.AddGraph(element.ModelElement, true);
     foundSome = true;
    }
   }
   if (!foundSome) return;

   // A DataObject carries a serialized version.
   System.Windows.Forms.IDataObject data =
                 new System.Windows.Forms.DataObject();
   data.SetData(elementGroup.CreatePrototype());
   System.Windows.Forms.Clipboard.SetDataObject
         (data,   // serialized clones of our selected model elements 
          false,  // we dont want to export outside this application
          10,     // retry 10 times on failure
          50);    // waiting 50ms between retries
  }

  #endregion

  #region Paste
  /// <summary>
  /// Add a command to Paste the content of model elements from the clipboard.
  /// </summary>
  /// <param name="commands">List of command to add the paste command to.</param>
  public void AddPasteCommand(List<MenuCommand> commands)
  {
   commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusPaste),
                                             new EventHandler(OnMenuPaste),
                                             StandardCommands.Paste));
  }

  /// <summary>
  /// Status for the Paste Command
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnStatusPaste(object sender, EventArgs e)
  {
   MenuCommand command = sender as MenuCommand;
   command.Visible = command.Enabled = true;
  }

  /// <summary>
  /// Paste command processing
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnMenuPaste(object sender, EventArgs e)
  {
   Diagram diagram = this.CurrentDocView.CurrentDiagram;
   if (diagram == null) return;

   System.Windows.Forms.IDataObject data =
                   System.Windows.Forms.Clipboard.GetDataObject();
   DesignSurfaceElementOperations op = diagram.ElementOperations;
   if (op.CanMerge(diagram, data))
   {
    // Find a suitable place to position the new shape.
    PointD place = new PointD(0, 0);
    foreach (object item in this.CurrentSelection)
    {
     ShapeElement shape = item as ShapeElement;
     if (shape != null)
     {
      place = shape.AbsoluteBoundingBox.Center;
      break;
     }
    }
    using (Transaction t = diagram.Store.
                  TransactionManager.BeginTransaction("paste"))
    {
     // Do the business.
     op.Merge(diagram, data, PointD.ToPointF(place));
     t.Commit();
    }
   }
  }
  #endregion

  #region LayoutDiagram
  /// <summary>
  /// Add a LayoutDiagram command
  /// </summary>
  /// <param name="commands">List of command to add the layout command to</param>
  public void AddLayoutDiagramCommand(List<MenuCommand> commands)
  {
   commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusLayoutDiagram),
                                             new EventHandler(OnMenuLayoutDiagram),
                                             new CommandID(guidDavDataCmdId, cmdidLayoutSelection)));
  }

  /// <summary>
  /// Status for the Paste Command
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnStatusLayoutDiagram(object sender, EventArgs e)
  {
   MenuCommand command = sender as MenuCommand;
   command.Visible = command.Enabled = true;
  }


  // From VSSDK\*\VisualStudioIntegration\Common\inc\stdidcmd.h
  private static readonly Guid guidVSStd97 = new Guid("5efc7975-14bc-11cf-9b2b-00aa00573819");
  private static readonly Guid guidStandardCommandSet2K = new Guid("1496A755-94DE-11D0-8C3F-00C04FC2AAE2");
  private static readonly Guid guidDavDataCmdId = new Guid("732abe75-cd80-11d0-a2db-00aa00a3efff");
  private static readonly int cmdidLayoutSelection = 0x0003;

  /// <summary>
  /// Paste command processing
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnMenuLayoutDiagram(object sender, EventArgs e)
  {
   Diagram diagram = this.CurrentDocView.CurrentDiagram;
   if (diagram == null) return;

   using (Transaction t = diagram.Store.
                 TransactionManager.BeginTransaction("Layout diagram"))
   {
    if (this.CurrentDocumentSelection.Count == 0)
     diagram.AutoLayoutShapeElements(diagram.NestedChildShapes);
    else
     diagram.AutoLayoutShapeElements(this.CurrentDocumentSelection);

    if (t.HasPendingChanges)
     t.Commit();
   }
  }
  #endregion

 }
}
