using System;
using System.Collections.Generic;
using System.Text;
using EnvDTE;
using Microsoft.VisualStudio.Modeling;
using EnvDTE80;
using DslFactory.Utilities.DteCodeModel;
using System.Net.Security;
using System.ServiceModel;

namespace Dslfactory.wcfLanguage
{
/*
 public interface IReverseEngineering<T> where T : ModelElement
 {
  /// <summary>
  /// Update the model from the code 
  /// </summary>
  /// <param name="model">Root concept from the model (will be used to
  /// find the model elements specifically reverse engineered by the
  /// class implmenting this interface</param>
  void UpdateModelFromCode(T model);

  /// <summary>
  /// Can the model take the code type into account ?
  /// </summary>
  /// <param name="codeType">Code type to decide if this Reverse engineerer
  /// can take it into account</param>
  bool CanTakeIntoAccount(CodeType codeType, T model);

  /// <summary>
  /// Takes the Code type into account if it can do it.
  /// </summary>
  /// <param name="codeType">Code type to take into account in the model</param>
  void TakeIntoAccount(CodeType codeType, T model);

  /// <summary>
  /// Can the reverse engineerer take into account the fact that an attribute was
  /// added or changed ?
  /// </summary>
  /// <param name="attribute">Attribute to take into account</param>
  bool CanProcessAttributeAddedOrChanged(CodeAttribute2 attribute, T model);

  /// <summary>
  /// Processes the fact that an attribute was added or changed ?
  /// </summary>
  /// <param name="attribute"></param>
  void ProcessAttributeAddedOrChanged(CodeAttribute2 attribute, T model);

  /// <summary>
  /// Can the reverse engineerer take into account the fact that an attribute was
  /// removed ?
  /// </summary>
  /// <param name="parent">Code element holding the attribute that was removed</param>
  /// <param name="attributeName"></param>
  bool CanProcessAttributeRemoved(CodeElement parent, string attributeName, T model);

  /// <summary>
  ///  Processes the fact that an attribute was removed
  /// </summary>
  /// <param name="parent"></param>
  /// <param name="attributeName"></param>
  void ProcessAttributeRemoved(CodeElement parent, string attributeName, T model);
 }
*/

 internal class ReverseEngineering
 {
  /// <summary>
  /// Updates the model
  /// </summary>
  public static void UpdateModel(ServicesDefinition servicesDefinition)
  {
   KnownCodeTypes knownCodeTypes = KnownCodeTypes.FromStore(servicesDefinition.Store);
   if (knownCodeTypes != null)
   {
    // Update interfaces and OperationContract
    InterfaceReverseEngineering.UpdateInterfaces(servicesDefinition, knownCodeTypes);

    // Update message contracts
    MessageContractReverseEngineering.UpdateMessageContracts(servicesDefinition, knownCodeTypes);

    // Update classes and DataContracts
    DataContractReverseEngineering.UpdateDataContracts(servicesDefinition, knownCodeTypes);

    // Update services
    ServiceReverseEngineering.UpdateServices(servicesDefinition, knownCodeTypes);
   }
  }



  /// <summary>
  /// Can a CodeType be taken into account in the model ?
  /// </summary>
  /// <param name="codeType"></param>
  /// <returns></returns>
  internal static bool CanBeMerged(CodeType codeType)
  {
   // Interfaces that belong to the project can be dropped anyway
   if (codeType.InfoLocation == vsCMInfoLocation.vsCMInfoLocationProject)
   {
    return true;
   }

   // External to project types
   else
   {
    // We can reverse engineer Interfaces provided they have a [InterfaceContract] attribute
    if (MessageContractReverseEngineering.CanMerge(codeType))
     return true;
    else if (DataContractReverseEngineering.CanMerge(codeType))
     return true;
    else if (InterfaceReverseEngineering.CanMerge(codeType))
     return true;
    else if (ServiceReverseEngineering.CanMerge(codeType))
     return true;
   }
   return false;
  }

  /// <summary>
  /// Takes a CodeType into account in the model
  /// </summary>
  /// <param name="ServicesDefinition"></param>
  /// <param name="codeType"></param>
  internal static TypeOfInterest Merge(ServicesDefinition ServicesDefinition, CodeType codeType)
  {
   InterfaceReverseEngineering.MergeType(ServicesDefinition, codeType);

   if (MessageContractReverseEngineering.CanMerge(codeType))
    return MessageContractReverseEngineering.MergeType(ServicesDefinition, codeType);
   else if (DataContractReverseEngineering.CanMerge(codeType))
    return DataContractReverseEngineering.MergeType(ServicesDefinition, codeType);
   else if (ServiceReverseEngineering.CanMerge(codeType))
    return ServiceReverseEngineering.MergeType(ServicesDefinition, codeType);
   // Simple class
   else
    return DataContractReverseEngineering.MergeType(ServicesDefinition, codeType);
 }

  /// <summary>
  /// Services definition
  /// </summary>
  ServicesDefinition servicesDefinition;

  /// <summary>
  /// Subscribe to Known types events
  /// </summary>
  /// <param name="servicesDefinition"></param>
  internal void SubscribeToKnownTypesEvents(ServicesDefinition servicesDefinition)
  {
   this.servicesDefinition = servicesDefinition;
   KnownCodeTypes knownTypeEvents = KnownCodeTypes.FromStore(servicesDefinition.Store);
   if (knownTypeEvents == null)
    return;

   knownTypeEvents.OnAttributeAdded += knownTypeEvents_OnAddedOrChanged;
   knownTypeEvents.OnAttributeChanged += knownTypeEvents_OnAddedOrChanged;
   knownTypeEvents.OnAttributeRemoved += knownTypeEvents_OnAttributeRemoved;
   knownTypeEvents.OnMethodAdded += knownTypeEvents_OnMethodAdded;
   knownTypeEvents.OnMethodRemoved += knownTypeEvents_OnMethodRemoved;
   knownTypeEvents.OnMethodRenamed += knownTypeEvents_OnMethodRenamed;
   knownTypeEvents.OnElementPropertiesChanged += knownTypeEvents_OnElementPropertiesChanged;
   
   knownTypeEvents.OnFieldAdded += knownTypeEvents_OnFieldAdded;
   knownTypeEvents.OnFieldRemoved += knownTypeEvents_OnFieldRemoved;
   knownTypeEvents.OnPropertyAdded += knownTypeEvents_OnPropertyAdded;
   knownTypeEvents.OnPropertyRemoved += knownTypeEvents_OnPropertyRemoved;

   servicesDefinition.Store.StoreDisposing += Store_StoreDisposing;
  }




  /// <summary>
  /// Unsubscribe to Store events when the Store is disposing.
  /// </summary>
  void Store_StoreDisposing(object sender, EventArgs e)
  {
   KnownCodeTypes knownTypeEvents = KnownCodeTypes.FromStore(servicesDefinition.Store);
   if (knownTypeEvents == null)
    return;

   knownTypeEvents.OnAttributeAdded -= knownTypeEvents_OnAddedOrChanged;
   knownTypeEvents.OnAttributeChanged -= knownTypeEvents_OnAddedOrChanged;
   knownTypeEvents.OnAttributeRemoved -= knownTypeEvents_OnAttributeRemoved;
   knownTypeEvents.OnMethodAdded -= knownTypeEvents_OnMethodAdded;
   knownTypeEvents.OnMethodRemoved -= knownTypeEvents_OnMethodRemoved;
   knownTypeEvents.OnMethodRenamed -= knownTypeEvents_OnMethodRenamed;
   knownTypeEvents.OnFieldAdded -= knownTypeEvents_OnFieldAdded;
   knownTypeEvents.OnFieldRemoved -= knownTypeEvents_OnFieldRemoved;
   knownTypeEvents.OnPropertyAdded -= knownTypeEvents_OnPropertyAdded;
   knownTypeEvents.OnPropertyRemoved -= knownTypeEvents_OnPropertyRemoved;
   knownTypeEvents.OnElementPropertiesChanged -= knownTypeEvents_OnElementPropertiesChanged;
   
  }

  /// <summary>
  /// Processes attribute changes
  /// </summary>
  /// <remarks>
  /// Attribute changes of interest are :
  /// - Attribute named "ServiceContractAttribute", on code interfaces, which makes the corresponding Model Interface
  ///   be updated
  /// - Attribute named "OperationContractAttribute" on code methods, which makes the corresponding Model Methods be updated
  /// </remarks>
  /// <param name="attribute">Changed attribute</param>
  void knownTypeEvents_OnAddedOrChanged(CodeAttribute2 attribute)
  {
   // Service contract attribute
   if (InterfaceReverseEngineering.CanProcessAttributeAddedOrChanged(attribute))
    InterfaceReverseEngineering.ProcessAttributeAddedOrChanged(servicesDefinition, attribute);

   // Message contract
   if (MessageContractReverseEngineering.CanProcessAttributeAddedOrChanged(attribute))
    MessageContractReverseEngineering.ProcessAttributeAddedOrChanged(servicesDefinition, attribute);

   // Data contract
   if (DataContractReverseEngineering.CanProcessAttributeAddedOrChanged(attribute))
    DataContractReverseEngineering.ProcessAttributeAddedOrChanged(servicesDefinition, attribute);

   if (ServiceReverseEngineering.CanProcessAttributeAddedOrChanged(attribute))
    ServiceReverseEngineering.ProcessAttributeAddedOrChanged(servicesDefinition, attribute);
  }

  void knownTypeEvents_OnAttributeRemoved(CodeElement parent, string attributeName)
  {
   // Service contract
   if (InterfaceReverseEngineering.IsServiceContractAttribute(attributeName))
    InterfaceReverseEngineering.ProcessServiceContractAttributeRemoved(parent, servicesDefinition);

   // Operation contract
   else if (InterfaceReverseEngineering.IsOperationContractAttribute(attributeName))
    InterfaceReverseEngineering.ProcessOperationContractAttributeRemoved(parent, servicesDefinition);

   // DataContract
   else if (DataContractReverseEngineering.IsDataContractAttribute(attributeName))
    DataContractReverseEngineering.ProcessDataContractAttributeRemoved(parent, servicesDefinition);

    // DataMember
   else if (DataContractReverseEngineering.IsDataMemberAttribute(attributeName))
    DataContractReverseEngineering.ProcessDataMemberAttributeRemoved(parent, servicesDefinition);

   else if (MessageContractReverseEngineering.IsMessageContractAttribute(attributeName))
    MessageContractReverseEngineering.ProcessMessageContractAttributeRemoved(parent, servicesDefinition);

   else if (    (MessageContractReverseEngineering.IsMessageHeaderAttributeFullName(attributeName)) 
             || (MessageContractReverseEngineering.IsMessageBodyMemberAttribute(attributeName)) 
           )
    MessageContractReverseEngineering.ProcessMessageMemberAttributeRemoved(parent, servicesDefinition);
  }


  void knownTypeEvents_OnMethodAdded(CodeType parent, CodeFunction2 newMethod)
  {
   InterfaceReverseEngineering.ProcessMethodAdded(parent, newMethod, servicesDefinition);
  }

  /// <summary>
  /// 
  /// </summary>
  /// <param name="type">Type on which a method was removed</param>
  /// <param name="methodUniqueSignature">in the sense of CodeFunction.get_Prototype((int)VsCMPrototypeUniqueSignature)</param>
  void knownTypeEvents_OnMethodRemoved(CodeType type, string methodUniqueSignature)
  {
   InterfaceReverseEngineering.ProcessMethodRemoved(type, methodUniqueSignature, servicesDefinition);
  }

  void knownTypeEvents_OnMethodRenamed(CodeType type, CodeFunction2 method)
  {
   UpdateModel(servicesDefinition);
  }

  void knownTypeEvents_OnElementPropertiesChanged(CodeElement element)
  {
   UpdateModel(servicesDefinition);
  }


  void knownTypeEvents_OnFieldAdded(CodeType parent, CodeVariable newField)
  {
   UpdateModel(servicesDefinition);
  }

  void knownTypeEvents_OnPropertyAdded(CodeType parent, CodeProperty newProperty)
  {
   UpdateModel(servicesDefinition);
  }

  void knownTypeEvents_OnPropertyRemoved(CodeType type, string memberNameOrSignature)
  {
   UpdateModel(servicesDefinition);
  }

  void knownTypeEvents_OnFieldRemoved(CodeType type, string memberNameOrSignature)
  {
   UpdateModel(servicesDefinition);
  }

 }
}
