using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel.Design;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Diagrams;
using Microsoft.VisualStudio.Modeling.Shell;
using Microsoft.VisualStudio.Shell;

namespace Dslfactory.wcfLanguage
{
 internal partial class wcfLanguageCommandSet
 {
  protected override IList<MenuCommand> GetMenuCommands()
  {
   // Use default commands
   List<MenuCommand> commands = new List<MenuCommand>(base.GetMenuCommands());

   // Add commands
   AddCopyCommand<TypeOfInterest>(commands);
   AddPasteCommand(commands);
   AddLayoutDiagramCommand(commands);
   AddGenerateCommand(commands);
   AddShowAsRoleCommand(commands);
   AddShowAsMemberCommand(commands);
   AddHideMemberCommand(commands);
   AddShowHiddenMembersCommand(commands);
   AddShowUsedConceptsCommand(commands);

   // That's it !
   return commands;
  }


  // Commands
  // --------
  #region Generate
  private void AddGenerateCommand(List<MenuCommand> commands)
  {
   commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusGenerate),
                                             new EventHandler(OnMenuGenerate),
                                             new CommandID(new Guid(Constants.wcfLanguageCommandSetId), 0x100)));
  }

  private void OnStatusGenerate(object sender, EventArgs e)
  {
   MenuCommand command = sender as MenuCommand;
   command.Visible = true;
   command.Enabled = false;

   foreach (object o in this.CurrentSelection)
   {
    // Pick out shapes representing Component model elements.
    ShapeElement shapeElement = o as ShapeElement;
    if (shapeElement != null)
    {
     TypeOfInterest modelElement = shapeElement.ModelElement as TypeOfInterest;
     if (modelElement != null)
      if (modelElement.GenerationKind == GenerationKind.NotGenerated)
      {
       command.Enabled = true;
       return;
      }
    }
   }
  }


  private void OnMenuGenerate(object sender, EventArgs e)
  {
   foreach (object o in this.CurrentSelection)
   {
    // Pick out shapes representing Component model elements.
    ShapeElement shapeElement = o as ShapeElement;
    TypeOfInterest modelElement = shapeElement.ModelElement as TypeOfInterest;
    if (modelElement != null)
     if (modelElement.GenerationKind == GenerationKind.NotGenerated)
      IncrementalCodeGeneration.Generate(modelElement);
   }
  }
  #endregion

  #region Copy
  /// <summary>
  /// Add the support for the standard Copy Command
  /// </summary>
  /// <param name="commands">CommandList, we want to add the Copy command to</param>
  /// <typeparam name="T">A ModelElement derived type which is used to restrict the kind of model element
  /// that can be copied to the clipboard</typeparam>
  public void AddCopyCommand<T>(List<MenuCommand> commands) where T : ModelElement
  {
   commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusCopy),
                                             new EventHandler(OnMenuCopy<T>),
                                             StandardCommands.Copy));
  }

  /// <summary>
  /// Status for the Copy command
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnStatusCopy(object sender, EventArgs e)
  {
   MenuCommand command = sender as MenuCommand;
   command.Visible = true;
   command.Enabled = this.CurrentSelection.Count > 0;
  }

  /// <summary>
  /// Copy command
  /// </summary>
  /// <typeparam name="T"></typeparam>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnMenuCopy<T>(object sender, EventArgs e) where T : ModelElement
  {
   Diagram diagram = this.CurrentDocView.CurrentDiagram;
   bool foundSome = false;
   ElementGroup elementGroup = new ElementGroup(diagram.Partition);
   foreach (object o in this.CurrentSelection)
   {
    // Pick out shapes representing Component model elements.
    ShapeElement element = o as ShapeElement;
    if (element != null && element.ModelElement != null
                        && element.ModelElement is T)
    {
     // add the element and its embedded children to the group
     elementGroup.AddGraph(element.ModelElement, true);
     foundSome = true;
    }
   }
   if (!foundSome) return;

   // A DataObject carries a serialized version.
   System.Windows.Forms.IDataObject data =
                 new System.Windows.Forms.DataObject();
   data.SetData(elementGroup.CreatePrototype());
   System.Windows.Forms.Clipboard.SetDataObject
         (data,   // serialized clones of our selected model elements 
          false,  // we dont want to export outside this application
          10,     // retry 10 times on failure
          50);    // waiting 50ms between retries
  }

  #endregion

  #region Paste
  public void AddPasteCommand(List<MenuCommand> commands)
  {
   commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusPaste),
                                             new EventHandler(OnMenuPaste),
                                             StandardCommands.Paste));
  }

  /// <summary>
  /// Status for the Paste Command
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnStatusPaste(object sender, EventArgs e)
  {
   MenuCommand command = sender as MenuCommand;
   command.Visible = command.Enabled = true;
  }

  /// <summary>
  /// Paste command processing
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnMenuPaste(object sender, EventArgs e)
  {
   Diagram diagram = this.CurrentDocView.CurrentDiagram;
   if (diagram == null) return;

   System.Windows.Forms.IDataObject data =
                   System.Windows.Forms.Clipboard.GetDataObject();
   DesignSurfaceElementOperations op = diagram.ElementOperations;
   if (op.CanMerge(diagram, data))
   {
    // Find a suitable place to position the new shape.
    PointD place = new PointD(0, 0);
    foreach (object item in this.CurrentSelection)
    {
     ShapeElement shape = item as ShapeElement;
     if (shape != null)
     {
      place = shape.AbsoluteBoundingBox.Center;
      break;
     }
    }
    using (Transaction t = diagram.Store.
                  TransactionManager.BeginTransaction("paste"))
    {
     // Do the business.
     op.Merge(diagram, data, PointD.ToPointF(place));
     t.Commit();
    }
   }
  }
  #endregion

  private static readonly Guid guidClassDesignerCommandSet = new Guid("FB41A027-57C5-4f83-9508-C326DCE6D943");
  
  #region LayoutDiagram
  public void AddLayoutDiagramCommand(List<MenuCommand> commands)
  {
   commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusLayoutDiagram),
                                             new EventHandler(OnMenuLayoutDiagram),
                                             new CommandID(guidClassDesignerCommandSet, 0x2889)));
  }

  /// <summary>
  /// Status for the Paste Command
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnStatusLayoutDiagram(object sender, EventArgs e)
  {
   MenuCommand command = sender as MenuCommand;
   command.Visible = command.Enabled = true;
  }

  /// <summary>
  /// Paste command processing
  /// </summary>
  /// <param name="sender"></param>
  /// <param name="e"></param>
  private void OnMenuLayoutDiagram(object sender, EventArgs e)
  {
   Diagram diagram = this.CurrentDocView.CurrentDiagram;
   if (diagram == null) return;

    using (Transaction t = diagram.Store.
                  TransactionManager.BeginTransaction("Layout diagram"))
    {
     if (this.CurrentDocumentSelection.Count == 0)
      diagram.AutoLayoutShapeElements(diagram.NestedChildShapes);
     else
      diagram.AutoLayoutShapeElements(this.CurrentDocumentSelection);

     if (t.HasPendingChanges)
      t.Commit();
    }
   }
  #endregion

  #region Show as role
   public void AddShowAsRoleCommand(List<MenuCommand> commands)
   {
    commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusShowAsRole),
                                              new EventHandler(OnMenuShowAsRole),
                                              new CommandID(new Guid(Constants.wcfLanguageCommandSetId), 0x101)));
   }

   /// <summary>
   /// Status for the Paste Command
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnStatusShowAsRole(object sender, EventArgs e)
   {
    MenuCommand command = sender as MenuCommand;
    command.Visible = command.Enabled = false;
    foreach(ModelElement mel in this.CurrentSelection)
     if (mel is FieldOrProperty)
     {
      FieldOrProperty dataMember = mel as FieldOrProperty;
      if (dataMember.Representation != RepresentationKind.AsRole)
       command.Visible = command.Enabled = true;
     }
   }

   /// <summary>
   /// Paste command processing
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
  private void OnMenuShowAsRole(object sender, EventArgs e)
   {
    Diagram diagram = this.CurrentDocView.CurrentDiagram;
    if (diagram == null) return;

    using (Transaction t = diagram.Store.
                  TransactionManager.BeginTransaction("Layout diagram"))
    {
     foreach (ModelElement mel in this.CurrentSelection)
      if (mel is FieldOrProperty)
      {
       FieldOrProperty dataMember = mel as FieldOrProperty;
       if (dataMember.Representation != RepresentationKind.AsRole)
        dataMember.Representation = RepresentationKind.AsRole;
      }

     if (t.HasPendingChanges)
      t.Commit();
    }
   }
   #endregion

  #region Show as member
   public void AddShowAsMemberCommand(List<MenuCommand> commands)
   {
    commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusShowAsMember),
                                              new EventHandler(OnMenuShowAsMember),
                                              new CommandID(new Guid(Constants.wcfLanguageCommandSetId), 0x102)));
   }

   /// <summary>
   /// Status for the Paste Command
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnStatusShowAsMember(object sender, EventArgs e)
   {
    MenuCommand command = sender as MenuCommand;
    command.Visible = command.Enabled = false;
    foreach (ModelElement mel in this.CurrentSelection)
     if (mel is UsageConnector)
     {
      DataClassUsesDataClasses usage = (mel as UsageConnector).ModelElement as DataClassUsesDataClasses;
      if (usage.RepresentedDataMember.Representation != RepresentationKind.AsMember)
       command.Visible = command.Enabled = true;
     }
   }

   /// <summary>
   /// Paste command processing
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnMenuShowAsMember(object sender, EventArgs e)
   {
    Diagram diagram = this.CurrentDocView.CurrentDiagram;
    if (diagram == null) return;

    using (Transaction t = diagram.Store.
                  TransactionManager.BeginTransaction("Show as member"))
    {

     foreach (ModelElement mel in this.CurrentSelection)
      if (mel is UsageConnector)
      {
       DataClassUsesDataClasses usage = (mel as UsageConnector).ModelElement as DataClassUsesDataClasses;
       if (usage.RepresentedDataMember.Representation != RepresentationKind.AsMember)
        usage.RepresentedDataMember.Representation = RepresentationKind.AsMember;
      }

     if (t.HasPendingChanges)
      t.Commit();
    }
   }
   #endregion

  #region Hide member
   public void AddHideMemberCommand(List<MenuCommand> commands)
   {
    commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusHideMember),
                                              new EventHandler(OnMenuHideMember),
                                              new CommandID(new Guid(Constants.wcfLanguageCommandSetId), 0x103)));
   }

   /// <summary>
   /// Status for the Paste Command
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnStatusHideMember(object sender, EventArgs e)
   {
    MenuCommand command = sender as MenuCommand;
    command.Visible = command.Enabled = false;
    foreach (ModelElement mel in this.CurrentSelection)
     if (mel is UsageConnector)
     {
      DataClassUsesDataClasses usage = (mel as UsageConnector).ModelElement as DataClassUsesDataClasses;
      if (usage.RepresentedDataMember.Representation != RepresentationKind.Hidden)
       command.Visible = command.Enabled = true;
     }
     else if (mel is FieldOrProperty)
     {
      FieldOrProperty dataMember = mel as FieldOrProperty;
      if (dataMember.Representation != RepresentationKind.Hidden)
       command.Visible = command.Enabled = true;
     }
   }

   /// <summary>
   /// Paste command processing
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnMenuHideMember(object sender, EventArgs e)
   {
    Diagram diagram = this.CurrentDocView.CurrentDiagram;
    if (diagram == null) return;

    using (Transaction t = diagram.Store.
                  TransactionManager.BeginTransaction("Hide member"))
    {

     foreach (ModelElement mel in this.CurrentSelection)
      if (mel is UsageConnector)
      {
       DataClassUsesDataClasses usage = (mel as UsageConnector).ModelElement as DataClassUsesDataClasses;
       if (usage.RepresentedDataMember.Representation != RepresentationKind.Hidden)
        usage.RepresentedDataMember.Representation = RepresentationKind.Hidden;
      } 
     else if (mel is FieldOrProperty)
     {
      FieldOrProperty dataMember = mel as FieldOrProperty;
      if (dataMember.Representation != RepresentationKind.Hidden)
       dataMember.Representation = RepresentationKind.Hidden;
     }



     if (t.HasPendingChanges)
      t.Commit();
    }
   }
   #endregion

  #region Show Hidden members
   public void AddShowHiddenMembersCommand(List<MenuCommand> commands)
   {
    commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusShowHiddenMembers),
                                              new EventHandler(OnMenuShowHiddenMembers),
                                              new CommandID(new Guid(Constants.wcfLanguageCommandSetId), 0x104)));
   }

   /// <summary>
   /// Status for the Paste Command
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnStatusShowHiddenMembers(object sender, EventArgs e)
   {
    MenuCommand command = sender as MenuCommand;
    command.Visible = command.Enabled = false;
    foreach (ModelElement mel in this.CurrentSelection)
     if (mel is DataClassShape)
      if ((mel as DataClassShape).IsFiltered)
      {
       command.Visible = command.Enabled = true;
       return;
      }
   }

   /// <summary>
   /// Paste command processing
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnMenuShowHiddenMembers(object sender, EventArgs e)
   {
    Diagram diagram = this.CurrentDocView.CurrentDiagram;
    if (diagram == null) return;

    using (Transaction t = diagram.Store.
                  TransactionManager.BeginTransaction("Show Hidden members"))
    {

     foreach (ModelElement mel in this.CurrentSelection)
      if (mel is DataClassShape)
      {
       DataClass dataContract = (mel as DataClassShape).ModelElement as DataClass;
       List<FieldOrProperty> members = new List<FieldOrProperty>(dataContract.Members);
       foreach (FieldOrProperty member in members)
        if (member.Representation == RepresentationKind.Hidden)
         member.Representation = RepresentationKind.AsMember;
      }

     if (t.HasPendingChanges)
      t.Commit();
    }
   }
   #endregion

   #region Show Used concepts
   public void AddShowUsedConceptsCommand(List<MenuCommand> commands)
   {
    commands.Add(new DynamicStatusMenuCommand(new EventHandler(OnStatusShowUsedConcepts),
                                              new EventHandler(OnMenuShowUsedConcepts),
                                              new CommandID(new Guid(Constants.wcfLanguageCommandSetId), 0x105)));
   }

   /// <summary>
   /// Status for the Paste Command
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnStatusShowUsedConcepts(object sender, EventArgs e)
   {
    MenuCommand command = sender as MenuCommand;
    command.Visible = command.Enabled = false;
    foreach (ModelElement mel in this.CurrentSelection)
     if ( (mel is TypeOfInterest) || ( (mel is ShapeElement) && ((mel as ShapeElement).ModelElement is TypeOfInterest) ) )
     {
      command.Visible = command.Enabled = true;
      return;
     }
   }

   /// <summary>
   /// Paste command processing
   /// </summary>
   /// <param name="sender"></param>
   /// <param name="e"></param>
   private void OnMenuShowUsedConcepts(object sender, EventArgs e)
   {
    Diagram diagram = this.CurrentDocView.CurrentDiagram;
    if (diagram == null) return;

    using (Transaction t = diagram.Store.
                  TransactionManager.BeginTransaction("Show Used concepts"))
    {

     foreach (ModelElement mel in this.CurrentSelection)
      if (mel is TypeOfInterest)
       (mel as TypeOfInterest).ShowUsedConcepts();
      else if ((mel is ShapeElement) && ((mel as ShapeElement).ModelElement is TypeOfInterest))
       ((mel as ShapeElement).ModelElement as TypeOfInterest).ShowUsedConcepts();

     if (t.HasPendingChanges)
      t.Commit();
    }
   }
   #endregion

 }
}
